@@ -1,11 +1,10 @@
 import uuid, time
 from muid.corpus import Corpus
+from hashlib import sha256
 
-def muid4( min_len=8, timeout=60*60 ):
-    return Memorable.muid(method=uuid.uuid4, min_len=min_len, timeout=timeout)
 
-def muid1( min_len=8, timeout=60*60 ):
-    return Memorable.muid(method=uuid.uuid1, min_len=min_len, timeout=timeout)
+def muid4( min_len=8, timeout=60*60, batch_size=30000 ):
+    return Memorable.muid4(min_len=min_len, timeout=timeout, batch_size=batch_size)
 
 def mhash(key):
     return Memorable.hash(key)
@@ -22,12 +21,29 @@ def mverify(key,min_len=8):
 class Memorable(Corpus):
 
     @staticmethod
-    def muid(min_len, timeout, method=None):
+    def default_key_generator(batch_size):
+        return [str(uuid.uuid4()) for _ in range(batch_size)]
+
+    @staticmethod
+    def muid(min_len, timeout, batch_size, method=None):
+        """ Generate a hash-memorable unique identifier """
         if method is None:
-            method = uuid.uuid4
-        gen = Memorable.key_generator(min_len=min_len, timeout=timeout, method=method)
+            method = Memorable.default_key_generator
+        gen = Memorable.key_generator(min_len=min_len, timeout=timeout, method=method, batch_size=batch_size)
         return next(gen)
 
+    @staticmethod
+    def hash(key):
+        return str(uuid.UUID(sha256(key.encode('utf-8')).hexdigest()[:32]))
+
+    @staticmethod
+    def alt_hash(key):
+        return str(uuid.uuid5(uuid.NAMESPACE_DNS, key))
+
+    @staticmethod
+    def muid4(min_len, timeout, batch_size):
+        return Memorable.muid(min_len=min_len, timeout=timeout, batch_size=batch_size, method=Memorable.default_key_generator)
+
     @staticmethod
     def mnemonic(key,separator=' ',capitalize=True):
         """ Reveal the memorable part of a key hash, if any """
@@ -35,8 +51,6 @@ class Memorable(Corpus):
         if result['result']:
             return Memorable.pretty(result['long'],separator=separator,capitalize=capitalize)
 
-
-
     @staticmethod
     def verify(key, min_len, verbose=False):
         if isinstance(key,str) and len(key):
@@ -48,10 +62,6 @@ class Memorable(Corpus):
             valid, long = False, ''
         return valid if not verbose else {"result":valid,"long":long}
 
-    @staticmethod
-    def hash(key):
-        return str(uuid.uuid5(uuid.NAMESPACE_DNS, key))
-
     @staticmethod
     def is_readable_hex(s):
         """ The set of strings refered to as henglish is the image of hex strings under the map which swaps (5->s,1->l,7->t,0->o) """
@@ -130,16 +140,19 @@ class Memorable(Corpus):
         cap_parts = [ part[0].upper()+part[1:] for part in parts ] if capitalize else parts
         return separator.join(cap_parts)
 
+
+
+
     @staticmethod
     def key_generator( method=None, min_len=7, timeout=5, verbose=False, batch_size=100 ):
         """ Returns generator that spits out valid keys whose hashes are recognizable Henglish words or word pairs
 
-             method:   function returning a unique identifier  (e.g.  uuid.uuid4)
+             method:   function returning a vector of unique identifiers
              timeout:  seconds
 
         """
         if method is None:
-            method = uuid.uuid4
+            method = Memorable.default_key_generator
         start_time = time.time()
         num_attempts = 0
 
@@ -147,24 +160,23 @@ class Memorable(Corpus):
         _min = Corpus.min_word_len()
         assert ( _min <= min_len <= _max )
         readable_phrases = Memorable.phrases_of_len(min_len) + Memorable.words_of_len(min_len)
-        hex_phrases = [ Memorable.from_readable_hex(readable) for readable in readable_phrases ]
+        hex_phrases = set( [ Memorable.from_readable_hex(readable) for readable in readable_phrases ] )
         while time.time()<start_time+timeout:
             num_attempts += batch_size
-            uuids    = [ method() for _ in range(batch_size) ]
-            codes    = [ Memorable.hash(str(uuid)) for uuid in uuids ]
-            patterns = [ code.replace('-','')[:min_len] for code in codes ]
-            found    = any( pattern in hex_phrases for pattern in patterns )
+            keys     = method(batch_size)
+            codes    = [ Memorable.hash(ky) for ky in keys ]
+            short_codes = set( [ code.replace('-','')[:min_len] for code in codes ] )
+            found    = any( short_code in hex_phrases for short_code in short_codes )
             if found:
-                for pattern, uid in zip( patterns, uuids ):
-                    if pattern in hex_phrases:
-                        key      = str(uid)
-                        code     = Memorable.hash(key)
-                        hcode    = Memorable.to_readable_hex(code=code)
-                        readable = Memorable.longest_word_or_phrase(hcode)
-                        pretty   = Memorable.pretty(readable, separator=' ', capitalize=True)
+                # Find again slowly...
+                short_codes = [code.replace('-', '')[:min_len] for code in codes]
+                for short_code, key in zip(short_codes,keys):
+                    if short_code in hex_phrases:
+                        short_readable = Memorable.to_readable_hex(short_code)
+                        pretty   = Memorable.pretty(short_readable, separator=' ', capitalize=True)
                         if verbose:
-                            yield {"length": len(readable), "pretty": pretty, "key": key, "hashed key": code,'num_attempts': num_attempts}
+                            yield {"length": len(short_readable), "pretty": pretty, "key": key, "hashed key": Memorable.hash(key),'num_attempts': num_attempts}
                         else:
                             yield key
-
+        yield {"timeout":True,"num_attempts":num_attempts,"key":None}
 
